package sun.font;

import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.AffineTransform;
import java.awt.geom.PathIterator;
import java.util.Iterator;
import java.util.ArrayList;
import java.awt.geom.GeneralPath;
import java.awt.Shape;
import java.awt.geom.Point2D;
import java.util.Formatter;
import java.awt.font.LayoutPath;

public abstract class LayoutPathImpl extends LayoutPath
{
    private static final boolean LOGMAP = false;
    private static final Formatter LOG;
    
    public LayoutPathImpl() {
    }
    
    public Point2D pointToPath(double x, double y) {
        Point2D.Double double1 = new Point2D.Double(x, y);
        super.pointToPath(double1, double1);
        return double1;
    }
    
    public Point2D pathToPoint(double x, double y, boolean b) {
        Point2D.Double double1 = new Point2D.Double(x, y);
        super.pathToPoint(double1, b, double1);
        return double1;
    }
    
    public void pointToPath(double n, double n2, Point2D point2D) {
        point2D.setLocation(n, n2);
        super.pointToPath(point2D, point2D);
    }
    
    public void pathToPoint(double n, double n2, boolean b, Point2D point2D) {
        point2D.setLocation(n, n2);
        super.pathToPoint(point2D, b, point2D);
    }
    
    public abstract double start();
    
    public abstract double end();
    
    public abstract double length();
    
    public abstract Shape mapShape(Shape p0);
    
    public static LayoutPathImpl getPath(LayoutPathImpl.EndType endType, double... array) {
        if ((array.length & 0x1) != 0x0) {
            throw new IllegalArgumentException("odd number of points not allowed");
        }
        return LayoutPathImpl.SegmentPath.get(endType, array);
    }
    
    static {
        LOG = new Formatter(System.out);
    }
    
    public enum EndType
    {
        PINNED, 
        EXTENDED, 
        CLOSED;
        
        public boolean isPinned() {
            return this == LayoutPathImpl.EndType.PINNED;
        }
        
        public boolean isExtended() {
            return this == LayoutPathImpl.EndType.EXTENDED;
        }
        
        public boolean isClosed() {
            return this == LayoutPathImpl.EndType.CLOSED;
        }
    }
    
    public static final class SegmentPathBuilder
    {
        private double[] data;
        private int w;
        private double px;
        private double py;
        private double a;
        private boolean pconnect;
        
        public void reset(int n) {
            if (this.data == null || n > this.data.length) {
                this.data = new double[n];
            }
            else if (n == 0) {
                this.data = null;
            }
            this.w = 0;
            double n2 = 0.0;
            this.py = n2;
            this.px = n2;
            this.pconnect = false;
        }
        
        public LayoutPathImpl.SegmentPath build(LayoutPathImpl.EndType endType, double... array) {
            assert array.length % 2 == 0;
            this.reset(array.length / 2 * 3);
            for (int i = 0; i < array.length; i += 2) {
                this.nextPoint(array[i], array[i + 1], i != 0);
            }
            return this.complete(endType);
        }
        
        public void moveTo(double n, double n2) {
            this.nextPoint(n, n2, false);
        }
        
        public void lineTo(double n, double n2) {
            this.nextPoint(n, n2, true);
        }
        
        private void nextPoint(double n, double n2, boolean pconnect) {
            if (n == this.px && n2 == this.py) {
                return;
            }
            if (this.w == 0) {
                if (this.data == null) {
                    this.data = new double[6];
                }
                if (pconnect) {
                    this.w = 3;
                }
            }
            if (this.w != 0 && !pconnect && !this.pconnect) {
                this.data[this.w - 3] = (this.px = n);
                this.data[this.w - 2] = (this.py = n2);
                return;
            }
            if (this.w == this.data.length) {
                double[] data = new double[this.w * 2];
                System.arraycopy(this.data, 0, data, 0, this.w);
                this.data = data;
            }
            if (pconnect) {
                double n3 = n - this.px;
                double n4 = n2 - this.py;
                this.a += Math.sqrt(n3 * n3 + n4 * n4);
            }
            this.data[this.w++] = n;
            this.data[this.w++] = n2;
            this.data[this.w++] = this.a;
            this.px = n;
            this.py = n2;
            this.pconnect = pconnect;
            return;
        }
        
        public LayoutPathImpl.SegmentPath complete() {
            return this.complete(LayoutPathImpl.EndType.EXTENDED);
        }
        
        public LayoutPathImpl.SegmentPath complete(LayoutPathImpl.EndType endType) {
            if (this.data == null || this.w < 6) {
                return null;
            }
            if (this.w == this.data.length) {
                LayoutPathImpl.SegmentPath segmentPath = new LayoutPathImpl.SegmentPath(this.data, endType);
                this.reset(0);
                return segmentPath;
            }
            double[] array = new double[this.w];
            System.arraycopy(this.data, 0, array, 0, this.w);
            LayoutPathImpl.SegmentPath segmentPath = new LayoutPathImpl.SegmentPath(array, endType);
            this.reset(2);
            return segmentPath;
        }
    }
    
    public static final class SegmentPath extends LayoutPathImpl
    {
        private double[] data;
        LayoutPathImpl.EndType etype;
        
        public static LayoutPathImpl.SegmentPath get(LayoutPathImpl.EndType endType, double... array) {
            return new LayoutPathImpl.SegmentPathBuilder().build(endType, array);
        }
        
        SegmentPath(double[] data, LayoutPathImpl.EndType etype) {
            this.data = data;
            this.etype = etype;
        }
        
        @Override
        public void pathToPoint(Point2D point2D, boolean b, Point2D point2D2) {
            this.locateAndGetIndex(point2D, b, point2D2);
        }
        
        @Override
        public boolean pointToPath(Point2D point2D, Point2D point2D2) {
            double x = point2D.getX();
            double y = point2D.getY();
            double n = this.data[0];
            double n2 = this.data[1];
            double n3 = this.data[2];
            double a = Double.MAX_VALUE;
            double n4 = 0.0;
            double n5 = 0.0;
            double n6 = 0.0;
            int n7 = 0;
            double n8;
            double n9;
            double n10;
            for (int i = 3; i < this.data.length; n = n8, n2 = n9, n3 = n10, i += 3) {
                n8 = this.data[i];
                n9 = this.data[i + 1];
                n10 = this.data[i + 2];
                double n11 = n8 - n;
                double n12 = n9 - n2;
                double n13 = n10 - n3;
                double n14 = n11 * (x - n) + n12 * (y - n2);
                if (n13 == 0.0 || (n14 < 0.0 && (!this.etype.isExtended() || i != 3))) {
                }
                else {
                    double n15 = n13 * n13;
                    if (n14 <= n15 || (this.etype.isExtended() && i == this.data.length - 3)) {
                        double n16 = n14 / n15;
                        n += n16 * n11;
                        n2 += n16 * n12;
                        n3 += n16 * n13;
                    }
                    else if (i == this.data.length - 3) {
                        n = n8;
                        n2 = n9;
                        n3 = n10;
                        i = this.data.length;
                    }
                    else {
                        continue;
                    }
                }
                double n17 = x - n;
                double n18 = y - n2;
                double n19 = n17 * n17 + n18 * n18;
                if (n19 <= a) {
                    a = n19;
                    n4 = n;
                    n5 = n2;
                    n6 = n3;
                    n7 = i;
                }
            }
            double n20 = this.data[n7 - 3];
            double n21 = this.data[n7 - 2];
            if (n4 != n20 || n5 != n21) {
                double n25 = this.data[n7];
                double n26 = this.data[n7 + 1];
                double sqrt = Math.sqrt(a);
                if ((x - n4) * (n26 - n21) > (y - n5) * (n25 - n20)) {
                    sqrt = -sqrt;
                }
                point2D2.setLocation(n6, sqrt);
                return false;
            }
            else {
                int n22 = n7 != 3 && this.data[n7 - 1] != this.data[n7 - 4];
                int n23 = n7 != this.data.length && this.data[n7 - 1] != this.data[n7 + 2];
                int n24 = this.etype.isExtended() && (n7 == 3 || n7 == this.data.length);
                if (n22 && n23) {
                    Point2D.Double location = new Point2D.Double(x, y);
                    this.calcoffset(n7 - 3, (boolean)n24, location);
                    Point2D.Double location2 = new Point2D.Double(x, y);
                    this.calcoffset(n7, (boolean)n24, location2);
                    if (Math.abs(location.y) > Math.abs(location2.y)) {
                        point2D2.setLocation(location);
                        return true;
                    }
                    point2D2.setLocation(location2);
                    return false;
                }
                else {
                    if (!(n22 == 0)) {
                        point2D2.setLocation(x, y);
                        this.calcoffset(n7 - 3, (boolean)n24, point2D2);
                        return true;
                    }
                    point2D2.setLocation(x, y);
                    this.calcoffset(n7, (boolean)n24, point2D2);
                    return false;
                }
            }
        }
        
        private void calcoffset(int n, boolean b, Point2D point2D) {
            double n2 = this.data[n - 3];
            double n3 = this.data[n - 2];
            double n4 = point2D.getX() - n2;
            double n5 = point2D.getY() - n3;
            double n6 = this.data[n] - n2;
            double n7 = this.data[n + 1] - n3;
            double n8 = this.data[n + 2] - this.data[n - 1];
            double n9 = (n4 * n6 + n5 * n7) / n8;
            double n10 = (n4 * -n7 + n5 * n6) / n8;
            if (!b) {
                if (n9 < 0.0) {
                    n9 = 0.0;
                }
                else if (n9 > n8) {
                    n9 = n8;
                }
            }
            point2D.setLocation(n9 + this.data[n - 1], n10);
        }
        
        @Override
        public Shape mapShape(Shape shape) {
            return new LayoutPathImpl.SegmentPath.Mapper().mapShape(shape);
        }
        
        @Override
        public double start() {
            return this.data[2];
        }
        
        @Override
        public double end() {
            return this.data[this.data.length - 1];
        }
        
        @Override
        public double length() {
            return this.data[this.data.length - 1] - this.data[2];
        }
        
        private double getClosedAdvance(double n, boolean b) {
            if (this.etype.isClosed()) {
                n -= this.data[2];
                n -= (double)(int)(n / this.length()) * this.length();
                if (n < 0.0 || (n == 0.0 && b)) {
                    n += this.length();
                }
                n += this.data[2];
            }
            return n;
        }
        
        private int getSegmentIndexForAdvance(double closedAdvance, boolean b) {
            closedAdvance = this.getClosedAdvance(closedAdvance, b);
            int i = 5;
            while (i < this.data.length - 1) {
                double n = this.data[i];
                if (closedAdvance >= n) {
                    if (closedAdvance == n && b) {
                        break;
                    }
                    i += 3;
                }
                else {
                    break;
                }
            }
            return i - 2;
        }
        
        private void map(int n, double n2, double n3, Point2D point2D) {
            double n4 = this.data[n] - this.data[n - 3];
            double n5 = this.data[n + 1] - this.data[n - 2];
            double n6 = this.data[n + 2] - this.data[n - 1];
            double n7 = n4 / n6;
            double n8 = n5 / n6;
            n2 -= this.data[n - 1];
            point2D.setLocation(this.data[n - 3] + n2 * n7 - n3 * n8, this.data[n - 2] + n2 * n8 + n3 * n7);
        }
        
        private int locateAndGetIndex(Point2D point2D, boolean b, Point2D point2D2) {
            double x = point2D.getX();
            double y = point2D.getY();
            int segmentIndexForAdvance = this.getSegmentIndexForAdvance(x, b);
            this.map(segmentIndexForAdvance, x, y, point2D2);
            return segmentIndexForAdvance;
        }
        
        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();
            sb.append("{");
            sb.append(this.etype.toString());
            sb.append(" ");
            for (int i = 0; i < this.data.length; i += 3) {
                if (i > 0) {
                    sb.append(",");
                }
                float f = (float)(int)(this.data[i] * 100.0) / 100.0f;
                float f2 = (float)(int)(this.data[i + 1] * 100.0) / 100.0f;
                float f3 = (float)(int)(this.data[i + 2] * 10.0) / 10.0f;
                sb.append("{");
                sb.append(f);
                sb.append(",");
                sb.append(f2);
                sb.append(",");
                sb.append(f3);
                sb.append("}");
            }
            sb.append("}");
            return sb.toString();
        }
        
        class LineInfo
        {
            double sx;
            double sy;
            double lx;
            double ly;
            double m;
            
            void set(double sx, double sy, double lx, double ly) {
                this.sx = sx;
                this.sy = sy;
                this.lx = lx;
                this.ly = ly;
                double n = lx - sx;
                if (n == 0.0) {
                    this.m = 0.0;
                }
                else {
                    this.m = (ly - sy) / n;
                }
            }
            
            void set(LayoutPathImpl.SegmentPath.LineInfo lineInfo) {
                this.sx = lineInfo.sx;
                this.sy = lineInfo.sy;
                this.lx = lineInfo.lx;
                this.ly = lineInfo.ly;
                this.m = lineInfo.m;
            }
            
            boolean pin(double n, double n2, LayoutPathImpl.SegmentPath.LineInfo lineInfo) {
                lineInfo.set(this);
                if (this.lx >= this.sx) {
                    if (this.sx < n2 && this.lx >= n) {
                        if (this.sx < n) {
                            if (this.m != 0.0) {
                                lineInfo.sy = this.sy + this.m * (n - this.sx);
                            }
                            lineInfo.sx = n;
                        }
                        if (this.lx > n2) {
                            if (this.m != 0.0) {
                                lineInfo.ly = this.ly + this.m * (n2 - this.lx);
                            }
                            lineInfo.lx = n2;
                        }
                        return true;
                    }
                }
                else if (this.lx < n2 && this.sx >= n) {
                    if (this.lx < n) {
                        if (this.m != 0.0) {
                            lineInfo.ly = this.ly + this.m * (n - this.lx);
                        }
                        lineInfo.lx = n;
                    }
                    if (this.sx > n2) {
                        if (this.m != 0.0) {
                            lineInfo.sy = this.sy + this.m * (n2 - this.sx);
                        }
                        lineInfo.sx = n2;
                    }
                    return true;
                }
                return false;
            }
            
            boolean pin(int n, LayoutPathImpl.SegmentPath.LineInfo lineInfo) {
                double n2 = LayoutPathImpl.SegmentPath.this.data[n - 1];
                double n3 = LayoutPathImpl.SegmentPath.this.data[n + 2];
                switch (LayoutPathImpl.SegmentPath.this.etype) {
                    case PINNED: {
                        break;
                    }
                    case EXTENDED: {
                        if (n == 3) {
                            n2 = Double.NEGATIVE_INFINITY;
                        }
                        if (n == LayoutPathImpl.SegmentPath.this.data.length - 3) {
                            n3 = Double.POSITIVE_INFINITY;
                            break;
                        }
                        break;
                    }
                }
                return this.pin(n2, n3, lineInfo);
            }
        }
        
        class Segment
        {
            final int ix;
            final double ux;
            final double uy;
            final LayoutPathImpl.SegmentPath.LineInfo temp;
            boolean broken;
            double cx;
            double cy;
            GeneralPath gp;
            
            Segment(int ix) {
                this.ix = ix;
                double n = LayoutPathImpl.SegmentPath.this.data[ix + 2] - LayoutPathImpl.SegmentPath.this.data[ix - 1];
                this.ux = (LayoutPathImpl.SegmentPath.this.data[ix] - LayoutPathImpl.SegmentPath.this.data[ix - 3]) / n;
                this.uy = (LayoutPathImpl.SegmentPath.this.data[ix + 1] - LayoutPathImpl.SegmentPath.this.data[ix - 2]) / n;
                this.temp = new LayoutPathImpl.SegmentPath.LineInfo();
            }
            
            void init() {
                this.broken = true;
                double n = Double.MIN_VALUE;
                this.cy = n;
                this.cx = n;
                this.gp = new GeneralPath();
            }
            
            void move() {
                this.broken = true;
            }
            
            void close() {
                if (!this.broken) {
                    this.gp.closePath();
                }
            }
            
            void line(LayoutPathImpl.SegmentPath.LineInfo lineInfo) {
                if (lineInfo.pin(this.ix, this.temp)) {
                    LayoutPathImpl.SegmentPath.LineInfo temp = this.temp;
                    temp.sx -= LayoutPathImpl.SegmentPath.this.data[this.ix - 1];
                    double n = LayoutPathImpl.SegmentPath.this.data[this.ix - 3] + this.temp.sx * this.ux - this.temp.sy * this.uy;
                    double n2 = LayoutPathImpl.SegmentPath.this.data[this.ix - 2] + this.temp.sx * this.uy + this.temp.sy * this.ux;
                    LayoutPathImpl.SegmentPath.LineInfo temp2 = this.temp;
                    temp2.lx -= LayoutPathImpl.SegmentPath.this.data[this.ix - 1];
                    double cx = LayoutPathImpl.SegmentPath.this.data[this.ix - 3] + this.temp.lx * this.ux - this.temp.ly * this.uy;
                    double cy = LayoutPathImpl.SegmentPath.this.data[this.ix - 2] + this.temp.lx * this.uy + this.temp.ly * this.ux;
                    if (n != this.cx || n2 != this.cy) {
                        if (this.broken) {
                            this.gp.moveTo((float)n, (float)n2);
                        }
                        else {
                            this.gp.lineTo((float)n, (float)n2);
                        }
                    }
                    this.gp.lineTo((float)cx, (float)cy);
                    this.broken = false;
                    this.cx = cx;
                    this.cy = cy;
                }
            }
        }
        
        class Mapper
        {
            final LayoutPathImpl.SegmentPath.LineInfo li;
            final ArrayList<LayoutPathImpl.SegmentPath.Segment> segments;
            final Point2D.Double mpt;
            final Point2D.Double cpt;
            boolean haveMT;
            
            Mapper() {
                this.li = new LayoutPathImpl.SegmentPath.LineInfo();
                this.segments = new ArrayList();
                for (int i = 3; i < LayoutPathImpl.SegmentPath.this.data.length; i += 3) {
                    if (LayoutPathImpl.SegmentPath.this.data[i + 2] != LayoutPathImpl.SegmentPath.this.data[i - 1]) {
                        this.segments.add(new LayoutPathImpl.SegmentPath.Segment(i));
                    }
                }
                this.mpt = new Point2D.Double();
                this.cpt = new Point2D.Double();
            }
            
            void init() {
                this.haveMT = false;
                Iterator<LayoutPathImpl.SegmentPath.Segment> iterator = this.segments.iterator();
                while (iterator.hasNext()) {
                    ((LayoutPathImpl.SegmentPath.Segment)iterator.next()).init();
                }
            }
            
            void moveTo(double x, double y) {
                this.mpt.x = x;
                this.mpt.y = y;
                this.haveMT = true;
            }
            
            void lineTo(double x, double y) {
                if (this.haveMT) {
                    this.cpt.x = this.mpt.x;
                    this.cpt.y = this.mpt.y;
                }
                if (x == this.cpt.x && y == this.cpt.y) {
                    return;
                }
                if (this.haveMT) {
                    this.haveMT = false;
                    Iterator<LayoutPathImpl.SegmentPath.Segment> iterator = this.segments.iterator();
                    while (iterator.hasNext()) {
                        ((LayoutPathImpl.SegmentPath.Segment)iterator.next()).move();
                    }
                }
                this.li.set(this.cpt.x, this.cpt.y, x, y);
                Iterator<LayoutPathImpl.SegmentPath.Segment> iterator2 = this.segments.iterator();
                while (iterator2.hasNext()) {
                    ((LayoutPathImpl.SegmentPath.Segment)iterator2.next()).line(this.li);
                }
                this.cpt.x = x;
                this.cpt.y = y;
                return;
            }
            
            void close() {
                this.lineTo(this.mpt.x, this.mpt.y);
                Iterator<LayoutPathImpl.SegmentPath.Segment> iterator = this.segments.iterator();
                while (iterator.hasNext()) {
                    ((LayoutPathImpl.SegmentPath.Segment)iterator.next()).close();
                }
            }
            
            public Shape mapShape(Shape shape) {
                PathIterator pathIterator = shape.getPathIterator(null, 1.0);
                this.init();
                double[] array = new double[2];
                while (!pathIterator.isDone()) {
                    switch (pathIterator.currentSegment(array)) {
                        case 4: {
                            this.close();
                            break;
                        }
                        case 0: {
                            this.moveTo(array[0], array[1]);
                            break;
                        }
                        case 1: {
                            this.lineTo(array[0], array[1]);
                            break;
                        }
                    }
                    pathIterator.next();
                }
                GeneralPath generalPath = new GeneralPath();
                Iterator<LayoutPathImpl.SegmentPath.Segment> iterator = this.segments.iterator();
                while (iterator.hasNext()) {
                    generalPath.append(((LayoutPathImpl.SegmentPath.Segment)iterator.next()).gp, false);
                }
                return generalPath;
            }
        }
    }
    
    public static class EmptyPath extends LayoutPathImpl
    {
        private AffineTransform tx;
        
        public EmptyPath(AffineTransform tx) {
            this.tx = tx;
        }
        
        @Override
        public void pathToPoint(Point2D point2D, boolean b, Point2D ptDst) {
            if (this.tx != null) {
                this.tx.transform(point2D, ptDst);
            }
            else {
                ptDst.setLocation(point2D);
            }
        }
        
        @Override
        public boolean pointToPath(Point2D point2D, Point2D ptDst) {
            ptDst.setLocation(point2D);
            if (this.tx != null) {
                try {
                    this.tx.inverseTransform(point2D, ptDst);
                }
                catch (NoninvertibleTransformException ex) {}
            }
            return ptDst.getX() > 0.0;
        }
        
        @Override
        public double start() {
            return 0.0;
        }
        
        @Override
        public double end() {
            return 0.0;
        }
        
        @Override
        public double length() {
            return 0.0;
        }
        
        @Override
        public Shape mapShape(Shape pSrc) {
            if (this.tx != null) {
                return this.tx.createTransformedShape(pSrc);
            }
            return pSrc;
        }
    }
}
